/*
 *
 *  This file is part of
 *	MakeIndex - A formatter and format independent index processor
 *
 *  Copyright (C) 1989 by Chen & Harrison International Systems, Inc.
 *  Copyright (C) 1988 by Olivetti Research Center
 *  Copyright (C) 1987 by Regents of the University of California
 *
 *  Author:
 *	Pehong Chen
 *	Chen & Harrison International Systems, Inc.
 *	Palo Alto, California
 *	USA
 *	(phc@renoir.berkeley.edu or chen@orc.olivetti.com)
 *
 *  Contributors:
 *	Please refer to the CONTRIB file that comes with this release
 *	for a list of people who have contributed to this and/or previous
 *	release(s) of MakeIndex.
 *
 *  All rights reserved by the copyright holders.  See the copyright
 *  notice distributed with this software for a complete description of
 *  the conditions under which it is made available.
 *
 */

#include    "mkind.h"
#include    "scanst.h"

static line_number	sty_lc = 0;	/* line count */
static entry_number	sty_tc = 0;	/* total count */
entry_number		sty_ec = 0;	/* error count */

const char	*idx_keyword = IDX_KEYWORD;
char		idx_aopen = IDX_AOPEN;
char		idx_aclose = IDX_ACLOSE;
char		idx_level = IDX_LEVEL;
char		idx_ropen = IDX_ROPEN;
char		idx_rclose = IDX_RCLOSE;
char		idx_quote = IDX_QUOTE;
char		idx_actual = IDX_ACTUAL;
char		idx_encap = IDX_ENCAP;
char		idx_escape = IDX_ESCAPE;

const char	*preamble = PREAMBLE_DEF;
const char	*postamble = POSTAMBLE_DEF;
int		prelen = PREAMBLE_LEN;
int		postlen = POSTAMBLE_LEN;

const char	*setpage_open = SETPAGEOPEN_DEF;
const char	*setpage_close = SETPAGECLOSE_DEF;
int		setpagelen = SETPAGE_LEN;

const char	*group_skip = GROUPSKIP_DEF;
int		skiplen = GROUPSKIP_LEN;

int		headings_flag = HEADINGSFLAG_DEF;
const char	*heading_pre = HEADINGPRE_DEF;
const char	*heading_suf = HEADINGSUF_DEF;
int		headprelen = HEADINGPRE_LEN;
int		headsuflen = HEADINGSUF_LEN;

const char	*symhead_pos = SYMHEADPOS_DEF;
const char	*symhead_neg = SYMHEADNEG_DEF;

const char	*numhead_pos = NUMHEADPOS_DEF;
const char	*numhead_neg = NUMHEADNEG_DEF;

const char	*item_r[FIELD_MAX] = {ITEM0_DEF, ITEM1_DEF, ITEM2_DEF};
const char	*item_u[FIELD_MAX] = {"", ITEM1_DEF, ITEM2_DEF};
const char	*item_x[FIELD_MAX] = {"", ITEM1_DEF, ITEM2_DEF};

int		ilen_r[FIELD_MAX] = {ITEM_LEN, ITEM_LEN, ITEM_LEN};
int		ilen_u[FIELD_MAX] = {0, ITEM_LEN, ITEM_LEN};
int		ilen_x[FIELD_MAX] = {0, ITEM_LEN, ITEM_LEN};

const char	*delim_p[FIELD_MAX] = {DELIM_DEF, DELIM_DEF, DELIM_DEF};
const char	*delim_n = DELIM_DEF; /* page number separator */
const char	*delim_r = DELIMR_DEF;/* page range designator */
const char	*delim_t = DELIMT_DEF;/* page list terminating delimiter */

const char	*suffix_2p = "";	/* suffix for two page ranges */
const char	*suffix_3p = "";	/* suffix for three page ranges */
const char	*suffix_mp = "";	/* suffix for multiple page ranges */

const char	*encap_p = ENCAP0_DEF;/* encapsulator prefix */
const char	*encap_i = ENCAP1_DEF;/* encapsulator infix */
const char	*encap_s = ENCAP2_DEF;/* encapsulator postfix */

int		linelength_max = LINELENGTH_MAX;

int		indent_length = INDENTLEN_DEF;
const char	*indent_space = INDENTSPC_DEF;

const char	*page_comp = COMPOSITOR_DEF;

/* MakeIndex 2.13: Correct initialization of this array.  Previous
 * versions incorrectly overlapped ALPL and ARAB.  Because npg[] is
 * declared as short, we can only represent encoded page numbers in 0
 * .. 32767.  The encoding uses constant biases to represent different
 * types of page numbers in these ranges:
 *	roman-lower		    0 ..  9999
 *	roman-upper		10000 .. 19999
 *	arabic			20000 .. 29999
 *	alpha-lower		30000 .. 30026
 *	alpha-upper		30027 .. 30052
 * Negative page numbers would cause overlap into adjacent ranges, but
 * they are not recognized by the scan_xxx() functions in scanid.c.
 */
page_number	page_offset[PAGETYPE_MAX] =
{
/* ROML */ 0,
/* ROMU */ ROMAN_LOWER_OFFSET,
/* ARAB */ ROMAN_LOWER_OFFSET + ROMAN_UPPER_OFFSET,
/* ALPL */ ROMAN_LOWER_OFFSET + ROMAN_UPPER_OFFSET + ARABIC_OFFSET,
/* ALPU */ ROMAN_LOWER_OFFSET + ROMAN_UPPER_OFFSET + ARABIC_OFFSET +
	   ALPHA_LOWER_OFFSET,
};
static	const char	*page_prec = PRECEDENCE_DEF;
static	int		put_dot;

static	int	count_lfd ARGS((const char *str));
static  void	dump_chr ARGS((const char *specifier, char c));
static  void	dump_escaped_char ARGS((char c));
static  void	dump_int ARGS((const char *specifier, int n));
static  void	dump_str ARGS((const char *specifier, const char *value));
static	int	next_nonblank ARGS((void));
static	int	scan_char ARGS((char *c));
static	int	scan_spec ARGS((char *spec, size_t spec_max));
static	int	scan_string ARGS((const char **str));

void
dump_sty(VOID_ARG)
{

    (void)fprintf(stddbg, "%s%s\n",
		  "%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%",
		  "%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%");
    (void)fprintf(stddbg, "\n%%%% %s %s\n", pgm_fn, VERSION);

    (void)fprintf(stddbg, "\n%%%% Input style file specifiers\n\n");
    dump_chr(ACTUAL,idx_actual);
    dump_chr(ACLOSE,idx_aclose);
    dump_chr(AOPEN,idx_aopen);
    dump_chr(ENCAP,idx_encap);
    dump_chr(ESCAPE,idx_escape);
    dump_str(KEYWORD,idx_keyword);
    dump_chr(LEVEL,idx_level);
    dump_str(COMPOSITOR,page_comp);
    dump_chr(QUOTE,idx_quote);
    dump_chr(RCLOSE,idx_rclose);
    dump_chr(ROPEN,idx_ropen);

    (void)fprintf(stddbg, "\n%%%% Output style file specifiers\n\n");
    dump_str(DELIM_0,delim_p[0]);
    dump_str(DELIM_1,delim_p[1]);
    dump_str(DELIM_2,delim_p[2]);
    dump_str(DELIM_N,delim_n);
    dump_str(DELIM_R,delim_r);
    dump_str(DELIM_T,delim_t);
    dump_str(ENCAP_1,encap_i);
    dump_str(ENCAP_0,encap_p);
    dump_str(ENCAP_2,encap_s);
    dump_str(GROUP_SKIP,group_skip);
    dump_str(HEADING_PRE,heading_pre);
    dump_str(HEADING_SUF,heading_suf);
    dump_int(HEADINGS_FLAG,headings_flag);
    dump_int(INDENT_LENGTH,indent_length);
    dump_str(INDENT_SPACE,indent_space);
    dump_str(ITEM_0,item_r[0]);
    dump_str(ITEM_01,item_u[1]);
    dump_str(ITEM_1,item_r[1]);
    dump_str(ITEM_12,item_u[2]);
    dump_str(ITEM_2,item_r[2]);
    dump_str(ITEM_x1,item_x[1]);
    dump_str(ITEM_x2,item_x[2]);
    dump_int(LINEMAX,linelength_max);
    dump_str(NUMHEAD_NEG,numhead_neg);
    dump_str(NUMHEAD_POS,numhead_pos);
    dump_str(PRECEDENCE,page_prec);
    dump_str(POSTAMBLE,postamble);
    dump_str(PREAMBLE,preamble);
    dump_str(SETPAGEOPEN,setpage_open);
    dump_str(SETPAGECLOSE,setpage_close);
    dump_str(SUFFIX_2P,suffix_2p);
    dump_str(SUFFIX_3P,suffix_3p);
    dump_str(SUFFIX_MP,suffix_mp);
    dump_str(SYMHEAD_NEG,symhead_neg);
    dump_str(SYMHEAD_POS,symhead_pos);
    (void)fprintf(stddbg, "\n%s%s\n\n",
		  "%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%",
		  "%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%");
}

static void
#if STDC
dump_chr(const char *specifier, char c)
#else
dump_chr(specifier, c)
const char *specifier;
char c;
#endif
{
    (void)fprintf(stddbg, FMT_SPECIFIER, specifier);
    (void)fputc('\'', stddbg);
    dump_escaped_char(c);
    (void)fputc('\'', stddbg);
    (void)fputc('\n', stddbg);
}

static void
#if STDC
dump_escaped_char(char c)
#else
dump_escaped_char(c)
char c;
#endif
{
    switch (c)
    {
    case '\n':
	(void)fputc('\\', stddbg);
	(void)fputc('n', stddbg);
	break;

    case '\t':
	(void)fputc('\\', stddbg);
	(void)fputc('t', stddbg);
	break;

    case '\\':
    case '\"':
    case '\'':
	(void)fputc('\\', stddbg);
	(void)fputc(c, stddbg);
	break;

    default:
	(void)fputc(c, stddbg);
	break;
    }
}

static void
#if STDC
dump_int(const char *specifier, int n)
#else
dump_int(specifier, n)
const char *specifier;
int n;
#endif
{
    (void)fprintf(stddbg, FMT_SPECIFIER, specifier);
    (void)fprintf(stddbg, "%d\n", n);
}

static void
#if STDC
dump_str(const char *specifier, const char *value)
#else
dump_str (specifier, value)
const char *specifier;
const char *value;
#endif
{
    (void)fprintf(stddbg, FMT_SPECIFIER, specifier);
    (void)fputc('\"', stddbg);
    for ( ; *value ; ++value)
	dump_escaped_char(*value);
    (void)fputc('\"', stddbg);
    (void)fputc('\n', stddbg);
}

void
scan_sty(VOID_ARG)
{
    STATIC char spec[ARRAY_MAX];
    int	    tmp;

    MESSAGE("Scanning style file %s...", sty_fn);

    while (scan_spec(spec,sizeof(spec))) {
	sty_tc++;
	put_dot = TRUE;

	/* output pre- and post-ambles */
	if (STREQ(spec, PREAMBLE)) {
	    (void) scan_string(&preamble);
	    prelen = count_lfd(preamble);
	} else if (STREQ(spec, POSTAMBLE)) {
	    (void) scan_string(&postamble);
	    postlen = count_lfd(postamble);
	} else if (STREQ(spec, GROUP_SKIP)) {
	    (void) scan_string(&group_skip);
	    skiplen = count_lfd(group_skip);
	} else if (STREQ(spec, HEADINGS_FLAG)) {
	    SCAN_NO(&headings_flag);
	} else if (STREQ(spec, HEADING_PRE)) {
	    (void) scan_string(&heading_pre);
	    headprelen = count_lfd(heading_pre);
	} else if (STREQ(spec, HEADING_SUF)) {
	    (void) scan_string(&heading_suf);
	    headsuflen = count_lfd(heading_suf);
	} else if (STREQ(spec, SYMHEAD_POS)) {
	    (void) scan_string(&symhead_pos);
	} else if (STREQ(spec, SYMHEAD_NEG)) {
	    (void) scan_string(&symhead_neg);
	} else if (STREQ(spec, NUMHEAD_POS)) {
	    (void) scan_string(&numhead_pos);
	} else if (STREQ(spec, NUMHEAD_NEG)) {
	    (void) scan_string(&numhead_neg);
	} else if (STREQ(spec, SETPAGEOPEN)) {
	    (void) scan_string(&setpage_open);
	    setpagelen = count_lfd(setpage_open);
	} else if (STREQ(spec, SETPAGECLOSE)) {
	    (void) scan_string(&setpage_close);
	    setpagelen = count_lfd(setpage_close);
	    /* output index item commands */
	} else if (STREQ(spec, ITEM_0)) {
	    (void) scan_string((const char**)&item_r[0]);
	    ilen_r[0] = count_lfd(item_r[0]);
	} else if (STREQ(spec, ITEM_1)) {
	    (void) scan_string((const char**)&item_r[1]);
	    ilen_r[1] = count_lfd(item_r[1]);
	} else if (STREQ(spec, ITEM_2)) {
	    (void) scan_string((const char**)&item_r[2]);
	    ilen_r[2] = count_lfd(item_r[2]);
	} else if (STREQ(spec, ITEM_01)) {
	    (void) scan_string((const char**)&item_u[1]);
	    ilen_u[1] = count_lfd(item_u[1]);
	} else if (STREQ(spec, ITEM_12)) {
	    (void) scan_string((const char**)&item_u[2]);
	    ilen_u[2] = count_lfd(item_u[2]);
	} else if (STREQ(spec, ITEM_x1)) {
	    (void) scan_string((const char**)&item_x[1]);
	    ilen_x[1] = count_lfd(item_x[1]);
	} else if (STREQ(spec, ITEM_x2)) {
	    (void) scan_string((const char**)&item_x[2]);
	    ilen_x[2] = count_lfd(item_x[2]);
	    /* output encapsulators */
	} else if (STREQ(spec, ENCAP_0))
	    (void) scan_string(&encap_p);
	else if (STREQ(spec, ENCAP_1))
	    (void) scan_string(&encap_i);
	else if (STREQ(spec, ENCAP_2))
	    (void) scan_string(&encap_s);
	/* output delimiters */
	else if (STREQ(spec, DELIM_0))
	    (void) scan_string((const char**)&delim_p[0]);
	else if (STREQ(spec, DELIM_1))
	    (void) scan_string((const char**)&delim_p[1]);
	else if (STREQ(spec, DELIM_2))
	    (void) scan_string((const char**)&delim_p[2]);
	else if (STREQ(spec, DELIM_N))
	    (void) scan_string(&delim_n);
	else if (STREQ(spec, DELIM_R))
	    (void) scan_string(&delim_r);
	else if (STREQ(spec, DELIM_T))
	    (void) scan_string(&delim_t);
	else if (STREQ(spec, SUFFIX_2P))
	    (void) scan_string(&suffix_2p);
	else if (STREQ(spec, SUFFIX_3P))
	    (void) scan_string(&suffix_3p);
	else if (STREQ(spec, SUFFIX_MP))
	    (void) scan_string(&suffix_mp);
	/* output line width */
	else if (STREQ(spec, LINEMAX)) {
	    SCAN_NO(&tmp);
	    if (tmp > 0)
		linelength_max = tmp;
	    else
		STY_ERROR2("%s must be positive (got %d)",
			   LINEMAX, tmp);
	    /* output line indentation length */
	} else if (STREQ(spec, INDENT_LENGTH)) {
	    SCAN_NO(&tmp);
	    if (tmp >= 0)
		indent_length = tmp;
	    else
		STY_ERROR2("%s must be nonnegative (got %d)",
			   INDENT_LENGTH, tmp);
	    /* output line indentation */
	} else if (STREQ(spec, INDENT_SPACE)) {
	    (void) scan_string(&indent_space);
	    /* composite page delimiter */
	} else if (STREQ(spec, COMPOSITOR)) {
	    (void) scan_string(&page_comp);
	    /* page precedence */
	} else if (STREQ(spec, PRECEDENCE)) {
	    (void) scan_string(&page_prec);
	    process_precedence();
	    /* index input format */
	} else if (STREQ(spec, KEYWORD))
	    (void) scan_string(&idx_keyword);
	else if (STREQ(spec, AOPEN))
	    (void) scan_char(&idx_aopen);
	else if (STREQ(spec, ACLOSE))
	    (void) scan_char(&idx_aclose);
	else if (STREQ(spec, LEVEL))
	    (void) scan_char(&idx_level);
	else if (STREQ(spec, ROPEN))
	    (void) scan_char(&idx_ropen);
	else if (STREQ(spec, RCLOSE))
	    (void) scan_char(&idx_rclose);
	else if (STREQ(spec, QUOTE))
	{
	    (void) scan_char(&idx_quote);
	    if (german_sort && (idx_quote == '"'))
	    {
		STY_ERROR("In German sort mode, the quote character must be different from '\"'.\nQuote character restored to '%c'.",
			  IDX_QUOTE_GERMAN);
		idx_quote = IDX_QUOTE_GERMAN;
	    }
	}
	else if (STREQ(spec, ACTUAL))
	    (void) scan_char(&idx_actual);
	else if (STREQ(spec, ENCAP))
	    (void) scan_char(&idx_encap);
	else if (STREQ(spec, ESCAPE))
	    (void) scan_char(&idx_escape);
	else {
	    (void) next_nonblank();
	    STY_SKIPLINE;
	    STY_ERROR("Unknown specifier %s.\n", spec);
	    put_dot = FALSE;
	}
	if (put_dot) {
	    STY_DOT;
	}
    }

    /* check if quote and escape are distinct */
    if (idx_quote == idx_escape) {
	STY_ERROR(
	       "Quote and escape symbols must be distinct (both `%c' now).\n",
	       idx_quote);
	idx_quote = IDX_QUOTE;
	idx_escape = IDX_ESCAPE;
    }
    DONE(sty_tc - sty_ec, "attributes redefined", sty_ec, "ignored");
    CLOSE(sty_fp);
}

static int
#if STDC
scan_spec(char spec[], size_t spec_max)
#else
scan_spec(spec,spec_max)
char	spec[];
size_t	spec_max;
#endif
{
    size_t  i = 0;
    int	    c;

    while (TRUE)
    {
	if ((c = next_nonblank()) == -1)
	    return (FALSE);
	else if (c == COMMENT) {
	    STY_SKIPLINE;
	} else
	    break;
    }

    spec[i++] = TOLOWER(c);
    while ((i < spec_max) && ((c = GET_CHAR(sty_fp)) != SPC) &&
	   (c != TAB) && (c != LFD) && (c != EOF))
	spec[i++] = TOLOWER(c);
    if (i < spec_max) {
	spec[i] = NUL;
	if (c == EOF) {
	    STY_ERROR("No attribute for specifier %s (premature EOF)\n", spec);
	    return (-1);
	}
	if (c == LFD)
	    sty_lc++;
	return (TRUE);
    } else {
	spec[spec_max-1] = NUL;
	STY_ERROR2("Specifier %s too long (max %d).\n", spec, (int)spec_max);
	return (FALSE);
    }
}


static int
next_nonblank(VOID_ARG)
{
    int	    c;

    while (TRUE) {
	switch (c = GET_CHAR(sty_fp)) {
	case EOF:
	    return (-1);

	case LFD:
	    sty_lc++;
	    /* fall through */

	case SPC:
	case TAB:
	    break;

	default:
	    return (c);
	}
    }
#if    IBM_PC_TURBO
    return (-1);		/* not reached, but keeps compiler happy */
#endif
}


static int
#if STDC
scan_string(const char **str)
#else
scan_string(str)
const char	**str;
#endif
{
    STATIC char clone[ARRAY_MAX];
    int	    i = 0;
    int	    c;
    char    *p;

    switch (c = next_nonblank()) {
    case STR_DELIM:
	while (TRUE)
	{
	    switch (c = GET_CHAR(sty_fp)) {
	    case EOF:
		STY_ERROR("No closing delimiter in %s.\n", clone);
		return (FALSE);
	    case STR_DELIM:		/* found closing quotation mark */
		if (i < sizeof(clone))
		{
		    clone[i] = NUL;
		    p = (char *)chkalloc(i+1,sizeof(char));
		    strcpy(p, clone);
		    *str = (const char *)p;
		    return (TRUE);
		}
		else
		{
		    STY_SKIPLINE;
		    STY_ERROR2("Attribute string %s too long (max %u).\n",
			       clone, (unsigned int)sizeof(clone));
		    return (FALSE);
		}
	    case BSH:			/* found escaped character */
		switch (c = GET_CHAR(sty_fp))
		{
		case 'n':		/* \n -> newline */
		    c = LFD;
		    break;
		case 't':		/* \t -> tab */
		    c = TAB;
		    break;
		default:		/* \<other> -> <other> */
		    break;
		}
		if (i < sizeof(clone))
		    clone[i++] = (char)c;
		break;
	    default:
		if (c == LFD)
		    sty_lc++;
		if (i < sizeof(clone))
		    clone[i++] = (char) c;
		break;
	    }
	}
    case COMMENT:			/* ignore comments */
	STY_SKIPLINE;
	break;
    default:				/* unexpected input */
	STY_SKIPLINE;
	STY_ERROR("No opening delimiter.\n%s", "");
	return (FALSE);
    }
    return (TRUE);		       /* function value no longer used */
}


static int
#if STDC
scan_char(char *c)
#else
scan_char(c)
char   *c;
#endif
{
    int	    clone;

    /* Parse '<char>' or '\<escaped-char>', where <char> is anything
       other than a newline or single quote, and <escaped-char> is
       any character.  '\n' is treated as a newline, and '\t' as a
       horizontal tab. */

    for (;;)
    {
	switch (clone = next_nonblank())
	{
	case CHR_DELIM:
	    switch (clone = GET_CHAR(sty_fp))
	    {
	    case CHR_DELIM:		/* '' is illegal */
		STY_SKIPLINE;
		STY_ERROR("Premature closing delimiter.\n%s", "");
		return (FALSE);
	    case EOF:
		STY_ERROR("No character (premature EOF).\n%s", "");
		return (FALSE);
	    case LFD:			/* '<newline>' -> newline */
		ungetc(clone,sty_fp);
		/* fall through */
	    case BSH:			/* escaped character pair */
		clone = GET_CHAR(sty_fp);
		switch (clone)
		{
		case LFD:		/* '\<newline>' -> newline */
		    sty_lc++;
		    break;
		case 'n':		/* '\n' -> newline */
		    clone = '\n';
		    break;
		case 't':		/* '\t' -> tab */
		    clone = '\t';
		    break;
		default:		/* '\<other>' -> <other> */
		    break;
		}
		/* fall through */
	    default:			/* complete parse of '<something>' */
		if (GET_CHAR(sty_fp) == CHR_DELIM)
		{
		    *c = (char) clone;
		    return (TRUE);
		}
		else
		{
		    STY_ERROR("No closing quoted character delimiter, \
or too many letters.\n%s", "");
		    return (FALSE);
		}
	    }
	    /* break; */		/* NOT REACHED */
	case COMMENT:			/* ignore comment lines */
	    STY_SKIPLINE;
	    break;
	default:			/* found unexpected character */
	    STY_SKIPLINE;
	    STY_ERROR("No opening quoted character delimiter.\n%s", "");
	    return (FALSE);
	}
    }
#if    IBM_PC_TURBO
    return (FALSE);		/* not reached, but keeps compiler happy */
#endif
}


static int
#if STDC
count_lfd(const char *str)
#else
count_lfd(str)
const char   *str;
#endif
{
    int	    i = 0;
    int	    n = 0;

    while (str[i] != NUL) {
	if (str[i] == LFD)
	    n++;
	i++;
    }
    return (n);
}


void
process_precedence(VOID_ARG)
{
    page_number order[PAGETYPE_MAX];
    page_type	type[PAGETYPE_MAX];
    int     i = 0;
    int	    last;
    int	    roml = FALSE;
    int	    romu = FALSE;
    int	    arab = FALSE;
    int	    alpl = FALSE;
    int	    alpu = FALSE;

    /* check for legal specifiers first */
    while ((i < PAGETYPE_MAX) && (page_prec[i] != NUL)) {
	switch (page_prec[i]) {
	case ROMAN_LOWER:
	    if (roml) {
		MULTIPLE(ROMAN_LOWER);
	    } else
		roml = TRUE;
	    break;
	case ROMAN_UPPER:
	    if (romu) {
		MULTIPLE(ROMAN_UPPER);
	    } else
		romu = TRUE;
	    break;
	case ARABIC:
	    if (arab) {
		MULTIPLE(ARABIC);
	    } else
		arab = TRUE;
	    break;
	case ALPHA_LOWER:
	    if (alpl) {
		MULTIPLE(ALPHA_UPPER);
	    } else
		alpl = TRUE;
	    break;
	case ALPHA_UPPER:
	    if (alpu) {
		MULTIPLE(ALPHA_UPPER);
	    } else
		alpu = TRUE;
	    break;
	default:
	    STY_SKIPLINE;
	    STY_ERROR("Unknown type `%c' in page precedence specification.\n",
		      page_prec[i]);
	    return;
	}
	i++;
    }
    if (page_prec[i] != NUL) {
	STY_SKIPLINE;
	STY_ERROR("Page precedence specification string too long.\n%s", "");
	return;
    }
    last = i;
    switch (page_prec[0]) {
    case ROMAN_LOWER:
	order[0] = ROMAN_LOWER_OFFSET;
	type[0] = ROML;
	break;
    case ROMAN_UPPER:
	order[0] = ROMAN_UPPER_OFFSET;
	type[0] = ROMU;
	break;
    case ARABIC:
	order[0] = ARABIC_OFFSET;
	type[0] = ARAB;
	break;
    case ALPHA_LOWER:
	order[0] = ALPHA_LOWER_OFFSET;
	type[0] = ALPL;
	break;
    case ALPHA_UPPER:
	order[0] = ALPHA_LOWER_OFFSET;
	type[0] = ALPU;
	break;
    }

    i = 1;
    while (i < last) {
	switch (page_prec[i]) {
	case ROMAN_LOWER:
	    order[i] = order[i - 1] + ROMAN_LOWER_OFFSET;
	    type[i] = ROML;
	    break;
	case ROMAN_UPPER:
	    order[i] = order[i - 1] + ROMAN_UPPER_OFFSET;
	    type[i] = ROMU;
	    break;
	case ARABIC:
	    order[i] = order[i - 1] + ARABIC_OFFSET;
	    type[i] = ARAB;
	    break;
	case ALPHA_LOWER:
	    order[i] = order[i - 1] + ALPHA_LOWER_OFFSET;
	    type[i] = ALPL;
	    break;
	case ALPHA_UPPER:
	    order[i] = order[i - 1] + ALPHA_LOWER_OFFSET;
	    type[i] = ALPU;
	    break;
	}
	i++;
    }

    for (i = 0; i < PAGETYPE_MAX; i++) {
	page_offset[i] = PAGE_NUMBER_UNSET;
    }
    page_offset[type[0]] = 0;
    for (i = 1; i < last; i++) {
	page_offset[type[i]] = order[i - 1];
    }
    for (i = 0; i < PAGETYPE_MAX; i++) {
	if (page_offset[i] == PAGE_NUMBER_UNSET) {
	    switch (type[last - 1]) {
	    case ROML:
		order[last] = order[last - 1] + ROMAN_LOWER_OFFSET;
		break;
	    case ROMU:
		order[last] = order[last - 1] + ROMAN_UPPER_OFFSET;
		break;
	    case ARAB:
		order[last] = order[last - 1] + ARABIC_OFFSET;
		break;
	    case ALPL:
		order[last] = order[last - 1] + ALPHA_LOWER_OFFSET;
		break;
	    case ALPU:
		order[last] = order[last - 1] + ALPHA_UPPER_OFFSET;
		break;
	    }
	    type[last] = (page_type)i;
	    page_offset[i] = order[last];
	    last++;
	}
    }
}
